@@ -5,6 +5,7 @@ import socket
 import zlib
 import imp
 import subprocess as ssubprocess
+import shlex
 import sshuttle.helpers as helpers
 from sshuttle.helpers import debug2
 
@@ -62,20 +63,20 @@ def empackage(z, name, data=None):
 def connect(ssh_cmd, rhostport, python, stderr, options):
     portl = []
 
-    if (rhostport or '').count(':') > 1:
+    if re.sub(r'.*@', '', rhostport or '').count(':') > 1:
         if rhostport.count(']') or rhostport.count('['):
             result = rhostport.split(']')
             rhost = result[0].strip('[')
             if len(result) > 1:
                 result[1] = result[1].strip(':')
-                if result[1] is not '':
+                if result[1] != '':
                     portl = ['-p', str(int(result[1]))]
         # can't disambiguate IPv6 colons and a port number. pass the hostname
         # through.
         else:
             rhost = rhostport
     else:  # IPv4
-        l = (rhostport or '').split(':', 1)
+        l = (rhostport or '').rsplit(':', 1)
         rhost = l[0]
         if len(l) > 1:
             portl = ['-p', str(int(l[1]))]
@@ -109,13 +110,13 @@ def connect(ssh_cmd, rhostport, python, stderr, options):
         argv = [sys.executable, '-c', pyscript]
     else:
         if ssh_cmd:
-            sshl = ssh_cmd.split(' ')
+            sshl = shlex.split(ssh_cmd)
         else:
             sshl = ['ssh']
         if python:
             pycmd = "'%s' -c '%s'" % (python, pyscript)
         else:
-            pycmd = ("P=python3.5; $P -V 2>/dev/null || P=python; "
+            pycmd = ("P=python3; $P -V 2>/dev/null || P=python; "
                      "exec \"$P\" -c %s") % quote(pyscript)
             pycmd = ("exec /bin/sh -c %s" % quote(pycmd))
         argv = (sshl +
